import json
from typing import Optional

from pika.exceptions import AMQPError
from pika.exchange_type import ExchangeType
from tenacity import retry, retry_if_exception_type, wait_exponential
from typeguard import typechecked

from robotnikmq.config import RobotnikConfig
from robotnikmq.core import Robotnik, Message


class Topic(Robotnik):
    """The Topic is one of two key elements of the Publish/Subscribe workflow with RobotnikMQ.
       Once configured, a topic is able to broadcast messages to any Subscribers on a given exchange
       and routing key combination.
    """
    @typechecked
    def __init__(
        self,
        exchange: str,
        config: Optional[RobotnikConfig] = None,
    ):
        super().__init__(config=config)
        self.exchange = exchange
        self.channel.exchange_declare(
            exchange=self.exchange, exchange_type=ExchangeType.topic, auto_delete=True
        )

    @retry(
        retry=retry_if_exception_type((AMQPError, OSError)),
        wait=wait_exponential(multiplier=1, min=3, max=30),
    )
    @typechecked
    def broadcast(
        self,
        msg: Message,
        routing_key: Optional[str] = None
    ) -> None:
        """Broadcasts a message with an optional routing key.

        Args:
            msg (Message): The message to be broadcast.
            routing_key (Optional[str], optional): Routing key used to broadcast the message.
                                                   Defaults to an empty string.
            on_msg_error (AMQPErrorCallback, optional): _description_. Defaults to None.
        """
        msg = msg.with_routing_key(routing_key or msg.routing_key or "")
        self.log.info("Broadcasting message (routing-key: [{}]):\n{}",
                      msg.routing_key, json.dumps(msg.to_dict(), indent=4))
        self.channel.basic_publish(
            exchange=self.exchange,
            routing_key=msg.routing_key,
            body=json.dumps(msg.to_dict()),
        )
        self.log.success("Broadcast:\n{}", json.dumps(msg.to_dict(), indent=4))



from collections import namedtuple
from random import uniform
from time import sleep
from traceback import format_exc
from typing import Optional, Callable, List, Generator

from pika.exceptions import AMQPError
from pika.exchange_type import ExchangeType
from typeguard import typechecked

from robotnikmq.config import RobotnikConfig
from robotnikmq.core import Robotnik, Message
from robotnikmq.error import MalformedMessage

OnMessageCallback = Callable[[Message], None]

ExchangeBinding = namedtuple("ExchangeBinding", ["exchange", "binding_key"])


class Subscriber(Robotnik):
    TIMEOUT_MAX = 30
    TIMEOUT_MIN = 0.5
    TIMEOUT_STEP = 3
    TIMEOUT_JITTER = 2

    @typechecked
    def __init__(self, exchange_bindings: Optional[List[ExchangeBinding]] = None,
                 config: Optional[RobotnikConfig] = None,):
        super().__init__(config=config)
        self.exchange_bindings = exchange_bindings or []

    @typechecked
    def _bind(self, exchange_binding: ExchangeBinding) -> "Subscriber":
        self.exchange_bindings.append(exchange_binding)
        return self

    @typechecked
    def bind(self, exchange: str, binding_key: str = "#") -> "Subscriber":
        return self._bind(ExchangeBinding(exchange, binding_key))

    @typechecked
    def _consume(
        self, inactivity_timeout: Optional[float]
    ) -> Generator[Optional[Message], None, None]:
        with self.open_channel() as channel:
            queue_name = (
                channel.queue_declare(queue="", exclusive=True).method.queue or ""
            )
            for ex_b in self.exchange_bindings:
                channel.exchange_declare(
                    exchange=ex_b.exchange,
                    exchange_type=ExchangeType.topic,
                    auto_delete=True,
                )
                channel.queue_bind(
                    exchange=ex_b.exchange,
                    queue=queue_name,
                    routing_key=ex_b.binding_key,
                )
            try:
                for method, ___, body in channel.consume(
                    queue=queue_name,  # pragma: no cover
                    auto_ack=False,
                    inactivity_timeout=inactivity_timeout,
                ):
                    if method and body:
                        channel.basic_ack(delivery_tag=method.delivery_tag or 0)
                        try:
                            yield Message.of_json(body.decode())
                        except MalformedMessage:
                            self.log.debug(format_exc())
                    else:
                        yield None
            finally:
                try:
                    channel.cancel()
                    self.close_channel(channel)
                except AssertionError as exc:
                    self.log.warning(f"Unable to close channel: {exc}")
    
    @typechecked
    @staticmethod
    def _jitter(step: float, jitter: float) -> float:
        return uniform(step - jitter, step + jitter)
    
    @typechecked
    def _backoff_with_jitter(self, current_timeout: float,
                             timeout_min: Optional[float] = None,
                             timeout_step: Optional[float] = None,
                             timeout_jitter: Optional[float] = None,
                             timeout_max: Optional[float] = None) -> float:
        timeout_min = timeout_min or self.TIMEOUT_MIN
        timeout_step = timeout_step or self.TIMEOUT_STEP
        timeout_jitter = timeout_jitter or self.TIMEOUT_JITTER
        timeout_max = timeout_max or self.TIMEOUT_MAX
        self.log.warning(f"Backing off for {current_timeout} seconds before reconnecting...")
        sleep(current_timeout)
        self.log.warning("Reconnecting")
        return min(current_timeout + self._jitter(timeout_step, timeout_jitter), timeout_max)

    @typechecked
    def consume(
        self, inactivity_timeout: Optional[float] = None
    ) -> Generator[Optional[Message], None, None]:
        timeout = Subscriber.TIMEOUT_MIN
        while 42:
            try:
                for msg in self._consume(inactivity_timeout):
                    yield msg
                    timeout = Subscriber.TIMEOUT_MIN
            except (AMQPError, OSError) as exc:
                self.log.warning(f"Subscriber issue encountered: {exc}")
                timeout = self._backoff_with_jitter(timeout)



# pylint: disable=redefined-outer-name
from collections import defaultdict
from multiprocessing import current_process
from typing import Dict, NoReturn, Union, Optional, Any

from pika.exceptions import AMQPError  # type: ignore
from typeguard import typechecked

from robotnikmq.log import log
from robotnikmq.rpc_client import RpcClient
from robotnikmq.rpc_server import RpcServer

from tests.integration.utils import sub_process

try:
    from pytest_cov.embed import cleanup_on_sigterm  # type: ignore
except ImportError:
    pass
else:
    cleanup_on_sigterm()


def test_basic_rpc_call(robotnikmq_config):
    @typechecked
    def fib(index: int) -> int:
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        result = client.call("fibonacci", {"index": 10})
        assert result == 55


def test_basic_rpc_call_only_once(robotnikmq_config):
    @typechecked
    def fib(index: int) -> int:
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config, only_once=True)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server, terminate=False):
        client = RpcClient(robotnikmq_config)
        result = client.call("fibonacci", {"index": 10})
        assert result == 55


def test_basic_rpc_call_without_typehints(robotnikmq_config):
    def fib(index: int) -> int:
        """Returns the Nth fibonacci number"""
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        result = client.call("fibonacci", {"index": 10})
        assert result == 55


def test_basic_rpc_call_with_any(robotnikmq_config):
    def fib(index: Any) -> int:
        """Returns the Nth fibonacci number"""
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        result = client.call("fibonacci", {"index": 10})
        assert result == 55


def test_basic_rpc_call_docs(robotnikmq_config):
    @typechecked
    def fib(index: int, other: Union[int, float] = 0) -> int:
        """Returns the Nth fibonacci number"""
        if index == other:
            return int(other)
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        result = client.call("fibonacci.__doc__")
        log.debug(result)
        assert result["description"] == fib.__doc__
        assert result["inputs"] == {"index": "int", "other": "Union[int,float]"}
        assert result["rpc_queue"] == "fibonacci"


def test_basic_rpc_call_without_docs(robotnikmq_config):
    @typechecked
    def fib(index: int) -> int:
        """Returns the Nth fibonacci number"""
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib, register_docs=False)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        try:
            client.call("fibonacci.__doc__", timeout=0.5)
            assert False
        except TimeoutError as exc:
            assert str(exc).startswith("No response has been received for Request")


def test_multiserve_rpc_call(robotnikmq_config):
    @typechecked
    def _call1() -> str:
        return "call1"

    @typechecked
    def _call2() -> str:
        return "call2"

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("call1", _call1)
        serv.register_rpc("call2", _call2)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        result = client.call("call1")
        assert result == "call1"
        result = client.call("call2")
        assert result == "call2"


def test_faulty_rpc_call(robotnikmq_config):
    @typechecked
    def _faulty() -> NoReturn:
        raise RuntimeError("Shit happens!")

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("test", _faulty)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        result = client.call("test")
        assert result.details.startswith(
            "There was an error "
            "while processing the "
            "request, please refer "
            "to server log with "
            "request ID"
        )


def test_raise_error(robotnikmq_config):
    @typechecked
    def _faulty() -> NoReturn:
        raise RuntimeError("Shit happens!")

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("test", _faulty)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        try:
            client.call("test", raise_on_error=True)
            assert False
        except RuntimeError as exc:
            assert str(exc).startswith(
                "There was an error while processing the request"
            )


def test_type_safety(robotnikmq_config):
    @typechecked
    def fib(index: int) -> int:
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        try:
            client.call("fibonacci", {"index": 10.1}, raise_on_error=True)
            assert False
        except RuntimeError as exc:
            log.debug(str(exc))
            assert str(exc).startswith("Invalid type for index")


def test_arg_safety(robotnikmq_config):
    @typechecked
    def fib(index: int) -> int:
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        try:
            client.call("fibonacci", {"indexes": 10.1}, raise_on_error=True)
            assert False
        except RuntimeError as exc:
            log.debug(str(exc))
            assert str(exc).startswith("Missing required argument index")


def test_union_type_safety(robotnikmq_config):
    @typechecked
    def fib(index: Union[int, float]) -> int:
        index = int(index)
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        try:
            client.call("fibonacci", {"index": "10.1"}, raise_on_error=True)
            assert False
        except RuntimeError as exc:
            log.debug(str(exc))
            assert str(exc).startswith("Invalid type for index")


def test_union_rpc_call(robotnikmq_config):
    @typechecked
    def fib(index: Union[int, float]) -> int:
        index = int(index)
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        result = client.call("fibonacci", {"index": 10})
        assert result == 55


def test_default_rpc_call(robotnikmq_config):
    @typechecked
    def fib10(index: Union[int, float] = 10) -> int:
        index = int(index)
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib10(index - 1) + fib10(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib10)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        result = client.call("fibonacci", {})
        assert result == 55
        result = client.call("fibonacci", {"index": 10})
        assert result == 55
        result = client.call("fibonacci")
        assert result == 55


def test_optional_rpc_call(robotnikmq_config):
    @typechecked
    def fib(index: Optional[int] = None) -> int:
        index = index if index is not None else 10
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        result = client.call("fibonacci", {})
        assert result == 55
        result = client.call("fibonacci", {"index": 10})
        assert result == 55
        result = client.call("fibonacci")
        assert result == 55


def test_faulty_input_rpc_call(robotnikmq_config):
    @typechecked
    def fib(index: int) -> int:
        if index == 0:
            return 0
        if index == 1:
            return 1
        return fib(index - 1) + fib(index - 2)

    @typechecked
    def _fib(args: Dict[str, int]) -> Dict[str, int]:
        return {"result": fib(args["index"])}

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", _fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        result = client.call("fibonacci", str_args="invalid_json")
        assert result.details == ("Input could not be decoded as JSON")


def test_rpc_with_amqp_error(robotnikmq_config):
    @typechecked
    def fib(index: int) -> int:
        raise AMQPError("Shit happens with AMQP!")

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        try:
            result = client.call("fibonacci", {"index": 20}, timeout=0.5)
            log.debug(result)
            assert False
        except TimeoutError as exc:
            assert str(exc).startswith("No response has been received for Request")


def test_rpc_with_keyboard_interrupt(robotnikmq_config):
    @typechecked
    def fib(index: int) -> int:
        raise KeyboardInterrupt()

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("fibonacci", fib)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        try:
            result = client.call("fibonacci", {"index": 20}, timeout=0.5)
            log.debug(result)
            assert False
        except TimeoutError as exc:
            assert str(exc).startswith("No response has been received for Request")


def test_rpc_call_to_multiple_servers(robotnikmq_config):
    @typechecked
    def _process_name() -> Dict[str, str]:
        return {"process-name": current_process().name}

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("process_name", _process_name)
        serv.run()

    with sub_process(server, name="Server 1"):
        with sub_process(server, name="Server 2"):
            with sub_process(server, name="Server 3"):
                with sub_process(server, name="Server 4"):
                    with sub_process(server, name="Server 5"):
                        client = RpcClient(robotnikmq_config)
                        results = defaultdict(int)
                        for _ in range(100):
                            results[
                                client.call("process_name", timeout=0.5)["process-name"]
                            ] += 1
                        log.debug(results)
                        assert len(results) == 5


def test_dict_input(robotnikmq_config):
    @typechecked
    def _dictify(my_dict: Dict[str, str]) -> str:
        return f'{"".join([k + v for k, v in my_dict.items()])}'

    def server():
        serv = RpcServer(robotnikmq_config)
        serv.register_rpc("dictify", _dictify)
        serv.run()

    with sub_process(server):
        client = RpcClient(robotnikmq_config)
        assert (
            client.call("dictify", {"my_dict": {"stuff": "moar"}}, timeout=0.5)
            == "stuffmoar"
        )



from contextlib import contextmanager
from multiprocessing import Process
from time import sleep
from typing import Callable, Any, Tuple, Optional


USERNAME = "robotnik"
PASSWORD = "hackme"
VIRTUAL_HOST = "/robotnik"

META_QUEUE = "skynet.legion"


@contextmanager
def sub_process(
    target: Callable,
    args: Optional[Tuple[Any, ...]] = None,
    name: Optional[str] = None,
    terminate: bool = True,
):
    proc = Process(target=target, args=args or (), name=name)
    proc.start()
    try:
        sleep(0.2)
        yield proc
    finally:
        if terminate:
            proc.terminate()
        proc.join()



from arrow import get as to_arrow
from pytest import raises

from robotnikmq.core import Message


def test_message_init():
    msg = Message.of(contents={'stuff': 'something'})
    assert msg.contents['stuff'] == 'something'


def test_message_read_route():
    msg = Message.of(contents={'stuff': 'something'}, routing_key='test.route')
    assert msg.route == 'test.route'


def test_message_getitem():
    msg = Message.of(contents={'stuff': 'something'}, routing_key='test.route')
    assert msg['stuff'] == 'something'


def test_message_contains():
    msg = Message.of(contents={'stuff': 'something'}, routing_key='test.route')
    assert 'stuff' in msg


def test_message_keys():
    msg = Message.of(contents={'stuff': 'something'}, routing_key='test.route')
    assert set(msg.keys()) == {'stuff'}


def test_message_values():
    msg = Message.of(contents={'stuff': 'something'}, routing_key='test.route')
    assert set(msg.values()) == {'something'}


def test_message_iter():
    msg = Message.of(contents={'stuff': 'something'}, routing_key='test.route')
    assert list(i for i in msg) == [('stuff')]


def test_message_to_dict():
    msg = Message.of(contents={'stuff': 'something'})
    assert msg.to_dict()['timestamp'] == msg.timestamp.int_timestamp


def test_message_to_json():
    msg = Message.of(contents={'stuff': 'something'})
    print(msg.to_json())
    assert msg.to_json().startswith('{"routing_key": "", "contents": {"stuff": "something"}, "msg_')


def test_message_from_json():
    msg = Message.of_json('{"routing_key": "", "contents": {"stuff": "something"}, "msg_id": "11bcdb39-'
                     '56a1-4244-a567-0a5e8eaa125f", "timestamp": 1614471764}')
    assert msg.timestamp == to_arrow(1614471764)


def test_not_jsonable():
    with raises(ValueError):
        Message.of(contents={'set': {1, 2, 3, 4}})
